/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/module.h"

void * memmem( const void * haystack, size_t haystacklen, 
const void * needle, size_t needlelen);

#define GZIP "/bin/gzip"

int
gzip_is_present( void )
{
	int n;

	n = open( GZIP, O_RDONLY );
	if ( n < 0 ) {
		fatal( "%s : no such file, please install it..\n\n", GZIP );
		close( n );
		return( -1 );
	}
	close( n );
	return( 0 );
}

unsigned long
extract_image( char * compressed_file, char * uncompressed_file )
{
	FILE * pipe;
	char * begin = NULL;
	char * zone = NULL;
	char magic[] = { 0x1f, 0x8b, 0x08 };
	char * p_magic = NULL;
	char command[ BUFSIZ ];
	struct stat stat_s;
	unsigned long offset;
	unsigned long len;
	int fd_compressed_file;
	int fd_uncompressed_file;
	int n;

	offset = 0;

	n = gzip_is_present();
	if ( n < 0 ) {
		return( -1 );
	}

	fputs( "[!] try to get image from compressed imagefile..\n", 
		stdout );
	fd_compressed_file = open( compressed_file, O_RDONLY );
	if ( fd_compressed_file < 0 ) {
		fatal( "%s : no such kernel image file..\n\n", 
		compressed_file );
		return( -1 );
	}

	strcat( uncompressed_file, ".gz" );
	fd_uncompressed_file = open( uncompressed_file, 
	O_SYNC | O_RDWR | O_CREAT | O_EXCL, 0777 );
	if ( fd_uncompressed_file < 0 ) {
		fatal( "can't create file %s..\n\n", uncompressed_file );
		return( -1 );
	}

	n = fstat( fd_compressed_file, &stat_s );
	if ( n < 0 ) {
		fatal( "can't fstat()..\n" );
		return( -1 );
	}

	begin = zone = mmap( NULL, stat_s.st_size, PROT_READ, MAP_PRIVATE, 
	fd_compressed_file, 0 );
	for ( ;; ) {
		p_magic = memmem( zone, stat_s.st_size-offset-3, magic, 3 );
		if ( p_magic == NULL ) {
			perror( "extract_image:memmem()" );
			return( -1 );
		}
		fprintf( stdout, "[!] find zip magic string at: %p..\n", 
		p_magic );

		len = *( unsigned long * )( p_magic-4 );
		offset = ( unsigned long )( p_magic-begin );
		if ( (len+offset) == stat_s.st_size ) {
			fputs( "[!] write data to new image file..\n", stdout );
			write( fd_uncompressed_file, p_magic, len );
			break;
		}

		zone = p_magic + 3;
		offset += 3;
	}
		
	close( fd_compressed_file );
	close( fd_uncompressed_file );

	fputs( "[!] unzip new image file..\n", stdout );
	memset( command, '\0', sizeof(command)-1 );
	snprintf( command, sizeof(command), "%s -d %s", GZIP, 
	uncompressed_file );
	pipe = popen( command, "r" );
	if ( pipe == NULL ) {
		fatal( "unzip failed..\n" );
		return( -1 );
	}
	pclose( pipe );
	fputs( "[!] done..\n", stdout );
	return( 0 );
}

#define IMAGE_MAX 4096000

void
insert_module_in_image( char * kernel_image, char * mod, 
int n_syscall  )
{
	module_t module;
	char init_code[] = {
		"\xE8\x00\x00\x00\x00" /* call init_module */
		"\xC7\x05\x00\x00\x00\x00\x11\x11\x11\x11" /* restore syscall */
		"\xE8\x11\x11\x11\x11" /* call syscall */
		"\xC3" /* return */
	};
	char image[ IMAGE_MAX ];
	unsigned long bss_start;
	unsigned long bss_end;
	unsigned long new_bss_end;
	unsigned long bss_size;
	unsigned long init_addr;
	unsigned long module_size;
	unsigned long image_size;
	unsigned long mem;
	unsigned long off;
	unsigned long syscall_off;
	size_t p_syscall;
	int fd_module;
	int fd_image;
	int offset;
	int call;
	int end;
	int n;
	
	bss_start = get_symbol_by_name( "__bss_start" );
	if ( !bss_start ) {
		fatal( "can't find symbol __bss_start..\n\n" );
		return;
	}

	bss_end = get_symbol_by_name( "_end" );
	if ( !bss_end ) {
		fatal( "can't find symbol _end..\n\n" );
		return;
	}

	bss_size = bss_end - bss_start;

	fd_module = open( mod, O_RDONLY );
	if ( fd_module < 0 ) {
		fatal( "can't open %s..\n\n", module );
		return;
	}

	n = load_module( &module, fd_module );
	if ( n < 0 ) {
		return;
	}

	n = init_module( &module, bss_end+strlen(init_code) );
	if ( n < 0 ) {
		return;
	}

	init_addr = find_symaddr( &module, "init_module" );

	fd_image = open( kernel_image, O_RDONLY );
	if ( fd_image < 0 ) {
		fatal( "can't open kernel image %s..\n", kernel_image );
		return;
	}
	
	offset = 0;
	while ( (n = read(fd_image, image+offset, IMAGE_MAX-offset)) > 0 ) {
		offset += n;
	}
	close( fd_image );

	module_size = make_module( &module, 
	(unsigned long)image+offset+bss_size+strlen(init_code) );
	image_size = offset + bss_size + module_size + strlen( init_code );

	new_bss_end = bss_end + module_size + strlen( init_code );
	mem = (unsigned long)memmem( (void *)image, offset, &bss_end, 4 );
	if ( !mem ) {
		perror( "insert_module_in_image:memmem()" );
		return;
	}

	mem = (unsigned long)memmem( (void *)mem+4 , offset, 
	&bss_end, 4 );
	if ( !mem ) {
		perror( "insert_module_in_image:memmem()" );
		return;
	}

	memcpy( (void *)mem, &new_bss_end, 4 );

	off = p_sys_call_table - 0xc0100000;
	syscall_off = p_sys_call_table + 4 * n_syscall;
	memcpy( &p_syscall, (void *)(image+off+4*n_syscall), 4 );

	call = init_addr - ( bss_end + 5 );
	memcpy( init_code+1, &call, 4 );
	memcpy( init_code+7, &syscall_off, 4 );
	call = p_syscall - ( bss_end + 20 );
	memcpy( init_code+11, &p_syscall,4);
	memcpy( init_code+16, &call, 4 );

	memcpy( (void *)(image+off+4*n_syscall), &bss_end, 4 );
	memcpy( image+offset+bss_size, init_code, strlen(init_code) );

	fd_image = open( kernel_image, O_RDWR | O_CREAT | O_SYNC, 0777 );
	if ( fd_image < 0 ) {
		fatal( "can't open kernel image %S..\n", kernel_image );
		return;
	}

	end = image_size;
	while ( (n = write(fd_image, image+image_size-end, end)) > 0 ) {
		end -= n;
	}
	close( fd_image );	

	fprintf( stdout, "module %s inserted in\nkernel image %s and" 
		" syscall table modified\n\n", mod, kernel_image );
	return;
}

void
usage_bzimage( void )
{
    fputs( 
	"bzimage [ -ei ] <arguments>\n\n"
	"This command makes it possible to handle a compressed kernel\n"
	"image : get image from compressed file, patch it...\n\n" 
	"__options\n"
	"  -e [compressed kernel image] [image] : extract a image from a\n"
	"compressed file ;\n"
	"(e.g. -e /boot/bzImage /tmp/image-decompress)\n"
	"  -i [uncompressed image] [module] [syscall num] : insert\n"
	"[module] in static kernel image [uncompressed image] and\n"
	"a [syscall].\n\n", stdout );
    return;
}	

#define EXTRACT_IMAGE 0
#define INSERT_IMAGE 1

void
do_bzimage( int pargc, char * pargv[] )
{
	char c;
	int flag;

	NOARGUMENTS( pargc, 1, kmem );

	flag = -1;
	optind = 0;

	while ( (c = getopt(pargc, pargv, "ei")) != EOF ) {
		switch ( c ) {
		case 'e':
			flag = EXTRACT_IMAGE;
			break;
		case 'i':
			flag = INSERT_IMAGE;
			break;
		default:
			usage_kmem();
			return;
        }
    }	

    switch ( flag ) {
    case EXTRACT_IMAGE:
        if ( pargc != 4 ) {
            usage_bzimage();
            return;
        }

        extract_image( pargv[2], pargv[3] );
        break;

	case INSERT_IMAGE:
		if ( pargc != 5 ) {
			usage_bzimage();
			return;
		}

		insert_module_in_image( pargv[2], pargv[3], atoi(pargv[4]) );
		break;

    default:
        return;
    }

    return;
}

