/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <ctype.h>
#include <unistd.h>
#include "include/libcap.h"
#include "include/capability.h"
#include "include/kernsh.h"

/* ripped from libcap-1.10 and modified */
#define setbits( A, B ) _setbits( (__cap_s *)A, (__cap_s *)B ) 
static void 
_setbits( __cap_s * a, __cap_s * b )
{
    int n;
    for ( n = __CAP_BLKS; n--; )
    a->_blk[n] |= b->_blk[n];
}

#define clrbits( A, B ) _clrbits( (__cap_s *)A, (__cap_s *)B )
static void 
_clrbits( __cap_s * a, __cap_s * b )
{
    int n;
    for ( n = __CAP_BLKS; n--; )
    a->_blk[n] &= ~b->_blk[n];
}

static char const *
namcmp( char const * str, char const * nam )
{
	while ( *nam && tolower((unsigned char)*str) == *nam ) {
    	str++;
    	nam++;
    }

    if ( *nam || isalnum((unsigned char)*str) || *str == '_' ) {
    	return( NULL );
	}

    return( str );
}

static int 
lookupname( char const ** strp )
{
    char const * str = *strp;

    if ( isdigit(*str) ) {
    	unsigned long n = strtoul( str, (char **)&str, 0 );
    	if ( n >= __CAP_BITS ) return( -1 );
    	*strp = str;
    	return( n );
    } else {
    	char const * s;
    	int n;
    
		for ( n = __CAP_BITS; n--; ) {
        	if ( _cap_names[n] && (s = namcmp(str, _cap_names[n])) ) {
        		*strp = s;
        		return( n );
        	}
    	}
    	return( -1 );
	}
	return( -1 );
}

#define LIBCAP_EFF 01
#define LIBCAP_INH 02
#define LIBCAP_PER 04

/* convert an internal representation to a textual one. The textual
 * representation is stored in static memory. It will be overwritten
 * on the next occasion that this function is called.
 */
static int 
getstateflags( cap_t caps, int capno )
{
    int f = 0;

    if ( isset_cap((__cap_s *)(&caps->set.effective),capno) ) f |= LIBCAP_EFF;
    if ( isset_cap((__cap_s *)(&caps->set.inheritable),capno) ) f |= LIBCAP_INH;
    if ( isset_cap((__cap_s *)(&caps->set.permitted),capno) ) f |= LIBCAP_PER;

    return( f );
}

/* maximum output text length (16 per cap) */
#define CAP_TEXT_SIZE ( 16 * __CAP_BITS )

cap_t
cap_from_text( const char * str )
{
    cap_t res;
    __cap_s allones;
    int n;

    if ( str == NULL ) {
    	fatal( "bad argument\n" );
    	return( NULL );
    }

	res = cap_init();
    if ( !res ) return( NULL );
    for ( n = __CAP_BLKS; n--; ) allones._blk[n] = -1;

    for ( ;; ) {
    	char op;
    	int flags = 0;
		int listed = 0;
    	__cap_s list = { {0} };

		/* skip leading spaces */
		while ( isspace((unsigned char)*str) ) str++;
    	if ( !*str ) {
        	return( res );
		}

  		/* identify caps specified by this clause */
   		if ( isalnum((unsigned char)*str) || *str == '_' ) {
       		for ( ;; ) {
        		if ( namcmp(str, "all") ) {
            		str += 3;
            		list = allones;
        		} else {
            		n = lookupname( &str );
            		if ( n == -1 ) {
            			goto bad;
					}
            		list.raise_cap( n );
        		}
        		if ( *str != ',' ) break;
        		if ( !isalnum((unsigned char)*++str) && *str != '_' ) goto bad;
        	}
        	listed = 1;
    	} else if ( *str == '+' || *str == '-' ) {
			goto bad;  /* require a list of capabilities */
		} else {
        	list = allones;
		}

    	/* identify first operation on list of capabilities */
    	op = *str++;
    	if ( op == '=' && (*str == '+' || *str == '-') ) {
        	if ( !listed ) goto bad;
        	op = ( *str++ == '+' ? 'P' : 'M' ); /* skip '=' and take next op */
    	} else if ( op != '+' && op != '-' && op != '=' ) {
        	goto bad;
		}

    	/* cycle through list of actions */
    	do {
        	if ( *str && !isspace(*str) ) {
        		switch ( *str++ ) { /* Effective, Inheritable, Permitted */
		        	case 'e':
            			flags |= LIBCAP_EFF;
            			break;
        			case 'i':
            			flags |= LIBCAP_INH;
            			break;
        			case 'p':
            			flags |= LIBCAP_PER;
            			break;
        			default:
            			goto bad;
        		}
        	} else if ( op != '=' ) {
        		goto bad;
        	}

        	switch ( op ) { /* how do we interpret the caps? */
        	case '=':
       		case 'P': /* =+ */
        	case 'M': /* =- */
        		clrbits( &res->set.effective, &list );
        		clrbits( &res->set.inheritable, &list );
        		clrbits( &res->set.permitted, &list );
        		/* fall through */
        		if ( op == 'M' ) goto minus;
        	case '+':
        		if ( flags & LIBCAP_EFF ) setbits( &res->set.effective, &list );
        		if ( flags & LIBCAP_INH ) setbits( &res->set.inheritable, &list );
        		if ( flags & LIBCAP_PER ) setbits( &res->set.permitted, &list );
        		break;
        	case '-':
        	minus:
            	if ( flags & LIBCAP_EFF ) clrbits( &res->set.effective, &list );
        		if ( flags & LIBCAP_INH ) clrbits( &res->set.inheritable, &list );
        		if ( flags & LIBCAP_PER ) clrbits( &res->set.permitted, &list );
        		break;
        	}

        	/* new directive? */
        	if ( *str == '+' || *str == '-' ) {
        		if ( !listed ) {
            		goto bad;
        		}
        		flags = 0; /* reset the flags */
        		op = *str++;
        		if ( !isalpha(*str) ) goto bad;
        	}
    	} while ( *str && !isspace(*str) );
    }

bad:
    cap_free( &res );
    return( NULL );
}

#define CAP_TEXT_BUFFER_ZONE 100

char *
cap_to_text( cap_t caps, ssize_t * length_p )
{
    static char buf[ CAP_TEXT_SIZE+CAP_TEXT_BUFFER_ZONE ];
    char * p = NULL;
    int histo[ 8 ] = { 0 };
    int m, n, t;

    if ( !good_cap_t(caps) ) {
    	return( NULL );
    }

    for ( n = __CAP_BITS; n--; ) {
		histo[getstateflags(caps, n)]++;
	}

    for ( m=t=7; t--; ) {
    	if ( histo[t] > histo[m] ) m = t;
	}

    p = sprintf( buf, "=%s%s%s",
        (m & LIBCAP_EFF) ? "e" : "",
        (m & LIBCAP_INH) ? "i" : "",
        (m & LIBCAP_PER) ? "p" : "" ) + buf;

    for ( t = 8; t--; ) {
    	if ( t != m && histo[t] ) {
        	*p++ = ' ';
        	for ( n = 0; n != __CAP_BITS; n++ ) {
        		if ( getstateflags(caps, n) == t ) {
            		if ( _cap_names[n] ) p += sprintf( p, "%s,", _cap_names[n] );
            		else p += sprintf( p, "%d,", n );
            		if ( p - buf > CAP_TEXT_SIZE ) return( NULL );
            	}
			}
        	p--;
        	n = t & ~m;
        	if ( n ) {
        		p += sprintf( p, "+%s%s%s",
               		(n & LIBCAP_EFF) ? "e" : "",
               		(n & LIBCAP_INH) ? "i" : "",
               		(n & LIBCAP_PER) ? "p" : "" );
			}
        	n = ~t & m;
        	if ( n ) {
        		p += sprintf( p, "-%s%s%s",
               		(n & LIBCAP_EFF) ? "e" : "",
               		(n & LIBCAP_INH) ? "i" : "",
               		(n & LIBCAP_PER) ? "p" : "" );
			}
        	if ( p - buf > CAP_TEXT_SIZE ) {
        		return( NULL );
        	}
    	}
	}

    if ( length_p ) {
    	*length_p = p - buf;
    }
    return( _libcap_strdup(buf) );
}

int 
capgetp( pid_t pid, cap_t cap_d )
{
	int i;

	i = good_cap_t( cap_d );
    if ( !i ) {
    	return( -1 );
    }

    cap_d->head.pid = pid;
    i = capget( &cap_d->head, &cap_d->set );
    cap_d->head.version = _LINUX_CAPABILITY_VERSION;
    cap_d->head.pid = 0;

    return( i );
}

/* set the caps on a specific process/pg etc.. */
int 
capsetp( pid_t pid, cap_t cap_d )
{
    int error;

    if ( !good_cap_t(cap_d) ) {
		fatal( "Capabilities not good!\n" );
    	return( -1 );
    }

    cap_d->head.pid = pid;
    error = capset( &cap_d->head, &cap_d->set );
    cap_d->head.version = _LINUX_CAPABILITY_VERSION;
    cap_d->head.pid = 0;

    return( error );
}

/* this is an internal library function to duplicate a string and
 * tag the result as something cap_free can handle.
 */
char * 
_libcap_strdup(const char *old)
{
	__u32 * raw_data;

	if ( old == NULL ) {
		return( NULL );
	}

	raw_data = malloc( sizeof(__u32) + strlen(old) + 1 );
	if ( raw_data == NULL ) {
		perror( "_libcap_strdup:malloc()" );
		return( NULL );
	}

	*( raw_data++ ) = CAP_S_MAGIC;
	strcpy( (char *)raw_data, old );

	return( (char *)raw_data );
}

/* scrub and then liberate an internal capability set */
int 
cap_free( void * data_p )
{
    if ( good_cap_t(data_p) ) {
    	data_p = -1 + (__u32 *)data_p;
    	memset( data_p, 0, sizeof(__u32) + sizeof(struct _cap_struct) );
    	free( data_p );
    	data_p = NULL;
    	return( 0 );
    }

    if ( good_cap_string(data_p) ) {
    	int length = strlen( data_p ) + sizeof( __u32 );
        data_p = -1 + (__u32 *)data_p;
        memset( data_p, 0, length );
        free( data_p );
        data_p = NULL;
        return( 0 );
    }

    fatal( "don't recognize what we're supposed to liberate\n" );
    return( -1 );
}

/* obtain a blank set of capabilities */
cap_t 
cap_init( void )
{
    __u32 * raw_data;
    cap_t result;

    raw_data = malloc( sizeof(__u32)+sizeof(*result) );
    if ( raw_data == NULL ) {
		perror( "cap_init:malloc()" );
		return( NULL );
	}

    *raw_data = CAP_T_MAGIC;
    result = (cap_t)( raw_data+1 );
    memset( result, 0, sizeof(*result) );
    result->head.version = _LINUX_CAPABILITY_VERSION;

    return( result );
}

#define PROC_CAP "/proc/sys/kernel/cap-bound"
#define CAP_FULL_SET ( ~0 )
#define CAP_TO_MASK( flag ) ( 1 << (flag) )
#define cap_raise( cap, flag ) ( cap |= CAP_TO_MASK(flag) )
#define cap_lower( cap, flag ) ( cap &= ~CAP_TO_MASK(flag) )
#define cap_raised( cap, flag ) ( cap & CAP_TO_MASK(flag) & CAP_FULL_SET )

typedef __u32 kernel_cap_t;

int
get_current_caps( kernel_cap_t * caps )
{
	FILE * fp;

	fp = fopen( PROC_CAP, "r" );
	if ( fp == NULL ) {
		fatal( "can't open() file %s ..\n", PROC_CAP );
		return( -1 );
	}

	fscanf( fp, "%d", caps );
	fclose( fp );
	return( 0 );
}

int
get_number_caps( char * capability )
{
	int capnum = -1;

    if ( isdigit(capability[0]) ) {
        capnum = (int)strtol( capability, (char **)NULL, 10 );
        if ( capnum > __CAP_BITS ) {
            fatal( "invalid capability number %s\n", capability );
            return( -1 );
        }
    } else {
        int jndex = 0;

        while ( _cap_names[jndex] != NULL ) {
            if ( !strcmp(capability, _cap_names[jndex]) ) {
                capnum = jndex;
                break;
            }
            jndex++;
        }
    }

    if ( capnum < 0 ) {
        fatal( "%s: invalid capability\n", capability );
        return( -1 );
    }
	return( capnum );
}
	
void
listcaps( void )
{
	kernel_cap_t caps = 0;
	int i = 0;
	int n;
	
	n = get_current_caps( &caps );
	if ( n ) {
		return;
	}

	fprintf( stdout, "[#] Current capabilities : %#x\n", caps );
	fputs( "[#] List of capabilities :\n\n", stdout );
	while ( i < __CAP_BITS ) {
		fprintf( stdout, "[%d][%c] %s ;\n", i, 
			cap_raised(caps, i) ? '*' : ' ',
			_cap_names[i] );
		i++;
	}

	fputs( "\n* = capabilities currently allowed\n", stdout );
	return;
}

void
checkcaps( char * capability )
{
	kernel_cap_t caps = 0;
	int capnum;
	int n;

	n = get_current_caps( &caps );
	if ( n ) {
		return;
	}
	
	capnum = get_number_caps( capability );
	if ( capnum < 0 ) {
		return;
	}

	fprintf( stdout, "[%d][%s] %s\n", capnum,
		cap_raised(caps, capnum) ? "allowed" : "non allowed", 
		_cap_names[capnum] );
	return;
}

void
zerocaps( char * capability )
{
	FILE * fp;
	kernel_cap_t caps = 0;	
	int capnum;
	int n;

	n = get_current_caps( &caps );
	if ( n ) { 
		return;
	}

    capnum = get_number_caps( capability );
    if ( capnum < 0 ) {
		return;
	}

	cap_lower( caps, capnum );

    fp = fopen( PROC_CAP, "w" );
    if ( fp == NULL ) {
        fatal( "can't open() file %s ..\n", PROC_CAP );
        return;
    }

	fprintf( fp, "0x%x", caps );
	fclose( fp );

	return;
}

void
getcaps( pid_t pid )
{
    char * result = NULL;
    cap_t cap_d = NULL;
    ssize_t length;
    int i;

    cap_d = cap_init();
    if ( cap_d == NULL ) {
        fatal( "failed to make a blank capability set\n" );
        return;
    }

    i = capgetp( pid, cap_d );
    if ( i ) {
        fatal( "failed to get cap's for process %d\n", pid );
        return;
    }

    result = cap_to_text( cap_d, &length );
    fprintf( stdout, "capabilities for pid %d: %s\n", pid, result );
    cap_free( result );
    result = NULL;

    return;
}

void
setcaps( pid_t pid, char * capabilities )
{
    char * result = NULL;
    kernel_cap_t caps = 0;
    cap_t cap_d;
    ssize_t length;
    int capnum;
    int i;

    i = get_current_caps( &caps );
    if ( i ) {
        return;
    }

    capnum = get_number_caps( capabilities );
    if ( capnum < 0 ) {
        return;
    }

    i = cap_raised( caps, capnum );
    if ( !i ) {
        fatal( "capability cap_setpcap not set ..\n" );
        return;
    }

    cap_d = cap_from_text( capabilities );
    if ( cap_d == NULL ) {
        fatal( "failed to get capabilities from text\n\n" );
        return;
    }

    result = cap_to_text( cap_d, &length );
    fprintf( stdout, "[caps set to:\n%s\n]\n", result );
    cap_free( result );
    result = NULL;

    i = capsetp( pid, cap_d );
    if ( i ) {
        fatal( "failed to set cap's on process %d\n\n", pid );
		perror( "setcaps:capsetp" );
        return;
    }

    fprintf( stdout, "[caps set on %d]\n", pid );
    return;
}

void
usage_cap( void )
{
	fputs( "cap [ -lczgs ] <arguments>\n\n"
	"This command makes it possible to read and set the capabilities of\n"
	"a process running.\n\n"
	"__options\n"
	"  -l : list capabilities and their descriptions\n" 
	"(via /proc/sys/kernel/cap-bound) ;\n"
	"  -c [cap] : check if [cap] is set or not (via /proc/sys/kernel/cap-bound) ;\n"
	"  -z [cap] : zero out capability set (via /proc/sys/kernel/cap-bound) ;\n"
	"  -g [pid] : displays the capabilities on the queried process(es),\n" 
	"(via libcap library) the capabilities are displayed in the\n"
	"cap_to_text(3) format ;\n"
	"  -s [pid] [caps] : set the process capabilities of a running process\n"
	"(via libcap library) ;\n\n",
	stdout );
	return;
}

enum {
	SET_CAPS,
};

void
do_cap( int pargc, char * pargv[] )
{
	char c;
	int flag;

	flag = -1;
	optind = 0;
	
	NOARGUMENTS( pargc, 1, cap );
	while ( (c = getopt(pargc, pargv, "lsz:g:c:")) != EOF ) {
		switch ( c ) {
			case 'l':
				listcaps();
				break;
			case 'c':
				checkcaps( optarg );
				break;
			case 'g': 
				getcaps( atoi(optarg) );
				break;
			case 's':
				flag = SET_CAPS;
				break;
			case 'z':
				zerocaps( optarg );
				break;
			default:
				usage_cap();
				return;
		}
	}

	switch ( flag ) {
	case SET_CAPS:
		if ( pargc != 4 ) {
			usage_cap();
			return;
		}

		setcaps( atoi(pargv[2]), pargv[3] );
		break;

	default:
		return;
	}

	return;
}
