/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <linux/sys.h>
#include <sys/utsname.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "include/kernsh.h"
#include <readline/readline.h>
#include <readline/history.h>

#define VERSION "0.2b"

size_t p_sys_call_table;
size_t p_system_call;
unsigned long sys_call_table[ NR_syscalls ];
int file_descriptor_kmem;
int modules_not_supported = 0;
char * script_on_command_line = NULL;
char * path_system_map = NULL;
char * path_vmlinux = NULL;
char * path_file_kmem = NULL;
int created;		/* know if we create the /dev/kmem equivalent */
int mmap_use = 0;	/* define if we want to use mmap read/write functions */

void
print_kernsh_banner( void )
{
	fprintf( stdout, "-----------------------[ KERNSH %s (c) 2002 ]-----------------------\n\n",
		VERSION );
}

void
print_kernsh_usage( char * kernsh )
{
    fprintf( stderr, "[!] Usage: %s [options] <arguments>..\n",
        kernsh );
    fputs( "[!]\n", stderr );

    fputs( "[!] __options\n", stderr );
    fputs( "[!]    -h : this help\n", stderr );
    fputs( "[!]    -q : mode quiet\n", stderr );
    fputs( "[!]    -c <file> : submit a script to kernsh on "
        "command line\n", stderr );
    fputs( "[!]    -s <file> : path of System.map\n", stderr );
    fputs( "[!]    -v <file> : path of vmlinux\n", stderr );
    fputs( "[!]    -m : define if we want to use mmap read/write functions\n", 
		stderr );
    fputs( "[!]    -k <file> : replace /dev/kmem by your proper file\n", stderr );
    fputs( "[!]\n\n", stderr );
    exit( EXIT_FAILURE );
}

/* generator function for command completion.  STATE lets us know 
 * whether to start from scratch; without any state (i.e. STATE == 0), 
 * then we start at the top of the list. 
 * http://www.ugcs.caltech.edu/info/readline/rlman_2.html
 */
char *
command_generator( char * text, int state )
{
    char * name = NULL;
	static int list_index;
	static int len;

	/* if this is a new word to complete, initialize now.  This 
	 * includes saving the length of TEXT for efficiency, and 
	 * initializing the index variable to 0. 
	 */
	if ( !state ) {
		list_index = 0;
		len = strlen( text );
	}

	/* return the next name which partially matches from the 
	 * command list. 
	 */
	while ( (name = p_command_main[list_index].name) ) {
		list_index++;
		if ( strncmp(name, text, len) == 0 ) {
			return( dupstr(name) );
		}
	}

	/* if no names matched, then return NULL. */
	return( (char *)NULL );
}

/* attempt to complete on the contents of TEXT.  START and END show the
 * region of TEXT that contains the word to complete.  We can use the
 * entire line in case we want to do some simple parsing.  Return the
 * array of matches, or NULL if there aren't any. 
 * http://www.ugcs.caltech.edu/info/readline/rlman_2.html
 */
char **
fileman_completion( char * text, int start, int end )
{
	char ** matches = NULL;

	/* If this word is at the start of the line, then it is a command
	 * to complete.  Otherwise it is the name of a file in the current
	 * directory. 
	 */
	if ( start == 0 ) {
		matches = rl_completion_matches( text, 
			(rl_compentry_func_t *)command_generator );
	}

	return( matches );
}

void 
initialize_readline( void )
{
	/* allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "kernsh";

	/* tell the completer that we want a crack first. */
	rl_attempted_completion_function = 
		(CPPFunction *)fileman_completion;
	return;
}

void
end_session( void )
{
	int i;
	
	fputs( "\n", stdout );
	close_kmem( file_descriptor_kmem );
	
	if ( created != 0 ) {
		i = unlink( path_file_kmem );
		if ( i ) {
			perror( "end_session:unlink()" );
		}
	}

	exit( EXIT_SUCCESS );
}

void 
do_quit( int argc, char * argv[] )
{
	end_session();
}

#define ISBLANK( ch ) ( ((ch) == ' ') || ((ch) == '\t') )
int 
command( char * cmd )
{
	char command_name[ BUFSIZ ];
	char * p_command_name = NULL;
	char * argv[64];
	char * string = NULL;
	static const char delim[] = " \t\n";
	int argc = 0;
	int n;
	int i;
	int k;

	p_command_name = cmd;
	while ( *p_command_name && !ISBLANK(*p_command_name) ) {
		p_command_name++;
	}

	memcpy( command_name, cmd, p_command_name-cmd );
	command_name[p_command_name - cmd] = '\0';
	
	memset( argv, '\0', sizeof(argv) );
	argv[argc++] = p_command_name;
	string = (char *)strtok( p_command_name, delim );
	while ( string != NULL ) {
		argv[argc++] = string;
		string = (char *)strtok( NULL, delim );
	}
	argv[argc] = '\0';

	for ( i = 0; i < sizeof(p_command_main)-1; i++ ) {
		if ( p_command_main[i].name == NULL ) {
			break;
		}

		n = strcmp( p_command_main[i].name, command_name );
		if ( n == 0 ) {
			break;
		}
	}

	if ( p_command_main[i].name == "help" ){
		if ( argc == 1 ) {
			for ( i = 0; i < sizeof(p_command_main)-1; 
			i++ ) {
        			if ( p_command_main[i].name == NULL ) {
					break;
				}

				if ( p_command_main[i].name != 
				"module" || (p_command_main[i].name == "module" 
				&& !modules_not_supported) ) {
					fprintf( stdout, "%s\t: %s\n", 
					p_command_main[i].name, 
					p_command_main[i].description );
				}
			}
			fputs( "\nType \"help\" followed by command name "
			"for full documentation.\n\n", stdout );
			return( 0 );
		} else if ( argc == 2 ) {
			for ( k = 0; k < sizeof(p_command_main)-1; 
				k++ ) {
				if ( p_command_main[k].name == NULL ) {
					fatal( "undefined command..\n\n" );
					return( -1 );
				}

				if( !strcmp(p_command_main[k].name, argv[1])  ) {
					if(  p_command_main[k].usage != 0 ) {
						p_command_main[k].usage(); 
						return( 0 );
					} else {
						fatal( "no avaible help for this command\n\n" );
						return( -1 );
					}
				} else if ( strstr(p_command_main[k].name, argv[1]) ) {
					if ( p_command_main[k].usage != 0 ) {
						p_command_main[k].usage();
						return( 0 );
					}
				}
			}
		} else {
			fatal( "usage\n" );
			return( 0 );
		}
	}

	if ( p_command_main[i].name == NULL ) {
		fatal( "undefined command..\n\n" );
		return( -1 );
	}

	p_command_main[i].func( argc, argv );
	fputs( "\n", stdout );
	return( 0 );
}

void
initialization( int quiet )
{
	FILE * fd_file_system_map;
	struct utsname p_uname;
	struct stat file;
	int i;
	int n;

	if ( stat(path_file_kmem, &file) == 0 ) {
		created = 0;
	} else if ( errno == ENOENT ) {
		i = mknod( path_file_kmem, S_IFCHR|S_IRUSR|S_IWUSR, 256+2 );
		if ( i == -1 ) {
			if ( !quiet ) {
				perror( "initialization:mknod()" );
			}
			fatal( "can't create the special device (%s)..\n", path_file_kmem );
			exit( EXIT_FAILURE );
		} else {
			created = 1;
		}
	}

	if ( !quiet ) {
		uname( &p_uname );
		fprintf( stdout, "system name  : %s\n", p_uname.sysname );
		fprintf( stdout, "node name    : %s\n", p_uname.nodename );
		fprintf( stdout, "realease     : %s\n", p_uname.release );
		fprintf( stdout, "version      : %s\n", p_uname.version );
		fprintf( stdout, "machine name : %s\n", p_uname.machine );
		fprintf( stdout, "domain name  : %s\n", p_uname.__domainname );
		fprintf( stdout, "using mmap   : %d\n\n", mmap_use );
	}

	/* open /dev/kmem */
	file_descriptor_kmem = open_kmem( path_file_kmem );
	if ( file_descriptor_kmem < 0 ) {
		fatal( "can't open /dev/kmem..\n" );
		exit( -1 );
	}

	/* now the file descriptor is open so we can delete it ;) */
	if ( created != 0 ) {
		if ( unlink(path_file_kmem) == -1 ) {
			fatal( "can't remove %s, please do it by hand\n", path_file_kmem );
		} else {
			created = 0;
		}
	}

	/* check if System.map is present */
	fd_file_system_map = found_system_map();
	if ( fd_file_system_map == NULL ) {
		fatal( "[#] System.map is necessary to kernsh.\n"
		"[#] Run the script scripts/create_system_map.sh to create it and copy it in /boot\n"
		"[#] Run again kernsh\n\n" );
		exit( EXIT_FAILURE );
	}

	/* we're checking integrity of system_call and sys_call_table 
	 * before to continue.
	 */
	n = set_system_call();
	if ( n < 0 ) exit( EXIT_FAILURE );
	n = set_syscall_table();
	if ( n < 0 ) exit( EXIT_FAILURE );
	n = refresh_syscall_table();
	if ( n < 0 ) exit( EXIT_FAILURE );

	return;
}	

int 
main( int argc, char * argv[] )
{
	FILE * fpscript = NULL;
	char * line = NULL;
	char * s = NULL;
	char buffer[ 128 ];
	char c = 0;
	uid_t uid;
	int fd;
	int quiet;
	int i;
	
	/* test uid */
	uid = getuid(); 
	if ( uid ) { 
		fatal( "Hey ! You must be root\n" );
		return( -1 );
	}

	print_kernsh_banner();

	if ( argc == 2 && argv[1][0] != '\x2d' ) {
/*		command( argv[1] );*/
/*		return( -1 );*/
		c = (char)1;
	}

	quiet = mmap_use = 0;
	if ( c != 1 )
    while ( (c = getopt(argc, argv, "hqmc:s:v:k:")) != EOF ) {
        switch ( c ) {
            case 'q':
                quiet++;
                break;
            case 'c':
                script_on_command_line = optarg;
                break;
			case 'm':
				mmap_use = 1;
				break;
            case 's':
                path_system_map = optarg;
                break;
            case 'v':
                path_vmlinux = optarg;
                break;
            case 'h':
                print_kernsh_usage( argv[0] );
                break;
			case 'k':
				path_file_kmem = optarg;
				break;
            default:
                print_kernsh_usage( argv[0] );
                exit( EXIT_FAILURE );
        }
    }

	if ( path_file_kmem == NULL ) {
		path_file_kmem = KMEM;
	}

	initialization( quiet );
	initialize_readline();

	/* test if modules are supported */
#define KSYMS "/proc/modules"
	fd = open( KSYMS, O_RDONLY );
	if ( fd < 0 ) {
		fatal( "[!] Modules aren't supported..\n\n" );
		modules_not_supported++;
	}

	if ( script_on_command_line ) {
		fpscript = fopen( script_on_command_line, "r" );
		if ( !fpscript ) {
			perror( "main:fopen()" );
			exit( EXIT_FAILURE );
		}

		for ( i = 0 ; 
			fgets(buffer, sizeof(buffer), fpscript) != NULL ; i++ ) {
			buffer[ strlen(buffer)-1 ] = '\0';

			s = stripwhite( buffer );
			if ( *s ) {
				command( buffer );
			}
		}
		fclose( fpscript );
	} else if( c == 1) {
		command( argv[1] );
		return ( -1 ) ; /* strange :) */
	} else {
		for ( ;; ) {
			line = readline( PROMPT );		
			if ( !line ) {
				break;
			}

			line = parse_line( line );

			/* remove leading and trailing whitespace from the line.
		 	 * then, if there is anything left, add it to the history list
		 	 * and execute it. 
		 	 */
			s = stripwhite( line );
			if ( *s ) {
				add_history( s );
				command( s );
			}

			end_parse_line();
			free( line );
		}
	}

	end_session();
	return( -1 ); 
}
