/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <string.h>
#include "include/kernsh.h"
#include "include/module.h"

#define KMEM_FLAGS ( 0x20 + 0x10 + 0x40 + 0x80 + 0x100 )
#define KMALLOC_OBJ "modules/lkm-kmalloc.o"
#define KFREE_OBJ "modules/lkm-kfree.o"
#define STACK 0xc0000000
#define SYSCALL_KMALLOC 254
#define SYSCALL_KFREE 255

struct kma_struct {
	unsigned long ( *kexec )( size_t, int );
	int size;
	int flags;
	unsigned long mem;
};

struct kfr_struct {
	unsigned long ( *kexec )( const void * );
	const void * address;
};

int
open_kmem( char * path_file_kmem )
{
	return( open(path_file_kmem, O_RDWR, 0) );
}

void
close_kmem( int kd )
{
	close( kd );
}

size_t
round_the_offset( size_t offset )
{
	size_t oft = 0;

	oft = ( offset - (size_t)PAGE_OFFSET ) -
		( (offset-(size_t)PAGE_OFFSET) & ~4095 ); /* better ? & ~(getpagesize() - 1) */
	/* fprintf( stderr, "get offset: %d\n", oft ); */
	return( oft );
}

/* An alternative to writing/reading kmem. Original Idea by Guillaume
 * Pelat <pelat_g _ at _ epita.fr> modifying for match needs of kernsh. See
 * kmem_mmap.gz... 
 */
int
mmap_write_kmem( void * buf, size_t offset, int size, int kd )
{
	size_t oft = 0;
	char * mem = NULL;
	int i;
	
	/* first we need to round the offset to the x of 4096 due to the round in mmap */
	oft = round_the_offset( offset );

	mem = mmap( 0, oft+(size_t)size, PROT_WRITE, MAP_SHARED, kd, 
		((offset-(size_t)PAGE_OFFSET) & ~4095) );
	if ( mem == MAP_FAILED ) {
		fatal( "can't map memory in write mode !\n" );
		return( -1 );
	}

	memcpy( mem+oft, buf, size );

	/* this is not really needed because mumap do the same */
	i = msync( mem, oft+size, MS_SYNC | MS_INVALIDATE );
	if ( i == -1 ) {
		fatal( "can't write in the kernel !\n" );
		return( -1 );
	}
	i = munmap ( mem, oft+size );
	if ( i == -1 ) {
		fatal( "can't unmap the memory !\n" );
		return( -1 );
	}
	return( 0 );
}

int
mmap_read_kmem( void * buf, size_t offset, int size, int kd )
{
	size_t oft = 0;
	char * mem = NULL;
	int i;

	/* first we need to round due to the round in mmap*/
	oft = round_the_offset( offset );

	mem = mmap( 0, oft+(size_t)size, PROT_READ, MAP_SHARED, kd,
		((offset-(size_t)PAGE_OFFSET) & ~4095) );
	if ( mem == MAP_FAILED ) {
		fatal( "can't map memory in read mode !\n" );
		return( -1 );
	}

	memcpy( buf, mem+oft, size );
	
	i = munmap( mem, oft+size );
	if ( i == -1 ) {
		fatal( "can't unmap the memory !\n" );
		return( -1 );
	}
	return( 0 );
}

int
write_kmem( void * buf, size_t offset, int size, int kd )
{
	int n;

	if ( mmap_use == 1 && kd == file_descriptor_kmem ) {
		return( mmap_write_kmem(buf, offset, size, kd) );
	}

	n = lseek( kd, offset, SEEK_SET );
	if ( n != offset ) {
		perror( "write_kmem:lseek()" );
		return( -1 );
	}

	n = write( kd, buf, size );
	if ( n != size ) {
		perror( "write_kmem:write()" );
		return( -1 );
	}
	return( 0 );
}

int
read_kmem( void * buf, size_t offset, int size, int kd )
{
	int n;

	if ( mmap_use == 1 && kd == file_descriptor_kmem ) {
		return( mmap_read_kmem(buf, offset, size, kd) );
	}

	n = lseek( kd, offset, SEEK_SET );
	if ( n != offset ) {
		perror( "read_kmem:lseek()" );
		return( -1 );
	}
	
	n = read( kd, buf, size );
	if ( n < 0 ) {
		perror( "read_kmem:read()" );
		return( -1 );
	}
	return( 0 );
}

void
kdump( size_t addr, int ssize, int file_descriptor )
{
	char * c = NULL;
	int n;

	c = ( char * )malloc( ssize );
	if ( c == NULL ) {
		perror( "kdump:malloc()" );
		return;
	}

	memset( c, '\0', ssize );
	n = read_kmem( c, addr, ssize, file_descriptor );
	if ( n < 0 ) {
		return;
	}

	hexdump( c, ssize, addr );
	free( c );

	return;
}

/* Why disassemble the bzimage, when you can do it directly from hell ?
 * So you can track directly this f*** kernel horses ;)
 */
void
kdisass( size_t addr_s, int size, int file_descriptor )
{
	unsigned char * buf = NULL;
	int n = 0;

	buf = (unsigned char *)calloc( size, sizeof(char) );
	if ( buf == NULL ) {
		perror( "malloc()" );
		return;
	}

	n = read_kmem( buf, addr_s, size, file_descriptor );
	if ( n < 0 ) {
		return;
	}

	display_code_asm( addr_s, buf, size );
	free( buf );
	return;
}

size_t
khijack_syscall( int sysnr, size_t new_address, int fd )
{
	size_t save_address;
	int i;
	int n;

	for ( save_address = i = 0; i != NR_syscalls; i++ ) {
        if ( i == sysnr ) {
			save_address = sys_call_table[i];
			sys_call_table[i] = new_address;

			n = write_kmem( &sys_call_table, p_sys_call_table,
			sizeof(sys_call_table), fd );
			if ( n < 0 ) {
				fatal( "can't khijack() syscall, aborting !\n" );
				return( -1 );
			}
		}
	}
	return( save_address );
}

void
run_syscall_kmalloc( struct kma_struct * kmalloc, int nbr_syscall )
{
	long k;
	long __res;

	k = ( long )kmalloc;

	__asm__ volatile ( "int $0x80"
			: "=a" (__res)
			: "0" (nbr_syscall), "b" ((long)(k)) );

	return;
}

void
run_syscall_kfree( struct kfr_struct * kfree, int nbr_syscall )
{
	long k;
	long __res;

	k = ( long )kfree;

	__asm__ volatile ( "int $0x80"
			: "=a" (__res)
			: "0" (nbr_syscall), "b" ((long)(k)) );

	return;
}


int
build_structure_kmalloc( int sz, int nbr_syscall, 
unsigned long * address )
{
	struct kma_struct kmalloc;
	
	kmalloc.kexec = ( void * )get_symbol_by_name( "kmalloc" );
	if ( !kmalloc.kexec ) {
		fatal( "can't find symbol kmalloc..\n\n" );
		return( -1 );
	}

	kmalloc.size = sz;
	kmalloc.flags = KMEM_FLAGS;
	kmalloc.mem = 0;
	run_syscall_kmalloc( &kmalloc, nbr_syscall );
	if ( !kmalloc.mem ) {
		fatal( "kmalloc failed..\n" );
		return( -1 );
	}

   	fprintf( stdout, "memory malloc()ed at %#x\n", 
	(unsigned int)kmalloc.mem );
	*address = kmalloc.mem;
	return( 0 );
}

void
kmalloc( int ssize, unsigned long * address )
{
    char save[ BUFSIZ ];
    size_t addr_save;
	size_t m;
    int n;
    
    n = read_kmem( &save, STACK, 192, file_descriptor_kmem );
	if ( n < 0 ) return;
    n = kinsert( KMALLOC_OBJ, STACK, 0, 0 );
    if ( n < 0 ) return;

	addr_save = khijack_syscall( SYSCALL_KMALLOC, STACK, 
		file_descriptor_kmem );
	if ( addr_save > 0 ) {
		n = build_structure_kmalloc( ssize, SYSCALL_KMALLOC, address );
		if ( n < 0 ) return;
		m = khijack_syscall( SYSCALL_KMALLOC, addr_save, 
			file_descriptor_kmem );
		if ( m < 0 ) return;
		write_kmem( &save, STACK, 192, file_descriptor_kmem );
	}
    return;
}

int
build_structure_kfree( size_t address, int nbr_syscall )
{
	struct kfr_struct kfree;
	
	kfree.kexec = ( void * )get_symbol_by_name( "kfree" );
	if ( !kfree.kexec ) {
		fatal( "can't find symbol kfree..\n\n" );
		return( -1 );
	}

	kfree.address = (const void *)address;
	run_syscall_kfree( &kfree, nbr_syscall );
	fprintf( stdout, "memory to %#x free()ed\n\n", address );
	return( 0 );
}

void
kfree( size_t address )
{
    char save[ BUFSIZ ];
    size_t address_save;
    size_t m;
    int n;

    n = read_kmem( &save, STACK, 192, file_descriptor_kmem );
	if ( n < 0 ) return;
    n = kinsert( KFREE_OBJ, STACK, 0, 0 );
	if ( n < 0 ) return;
	address_save = khijack_syscall( SYSCALL_KFREE, STACK, 
		file_descriptor_kmem );
	if ( address_save < 0 ) return;
	n = build_structure_kfree( address, SYSCALL_KFREE );
	if ( n < 0 ) return;
	m = khijack_syscall( SYSCALL_KFREE, address_save, 
		file_descriptor_kmem );
	if ( m < 0 ) return;
	write_kmem( &save, STACK, 192, file_descriptor_kmem );

	return;
}

void 
kinsmod( char * module, char * function )
{
	unsigned long address;
	size_t func_addr;
	size_t address_save;
	size_t m;
	int len ;
	int lkm_file_desc;
	int n;

	lkm_file_desc = open( module, O_RDONLY );
	if ( lkm_file_desc < 0 ) {
		fatal( "%s : No such module..\n\n" );
		return;
	}

	len = module_size( lkm_file_desc );
	close( lkm_file_desc );
	if ( len < 0 ) {
		return;
	}

    kmalloc( len, &address );
    if ( address <= 0 ) {
		fatal( "can't kmalloc() memory.. aborting !\n" );
		return;
	}
	
    n = kinsert( module, (size_t)address, function, &func_addr );
    if ( n < 0 ) {
		return;
	}

	fputs( "\n", stdout );
    address_save = khijack_syscall( SYSCALL_KFREE, func_addr, 
	file_descriptor_kmem );
	if ( address_save < 0 ) {
		return;
	}

    call_syscall( SYSCALL_KFREE );
	m = khijack_syscall( SYSCALL_KFREE, address_save, 
	file_descriptor_kmem );
	if ( m < 0 ) {
		return;
	}
    fprintf( stdout, "module %s inserted into /dev/kmem at\n%#x "
	"and %s() executed !\n", module, (int)address, function );
	return;
}

void
krmmod( char * m, size_t addr, char * function )
{
	unsigned long func_addr = 0;
	size_t address_save, n;
	int len;
	void * module = NULL ;

	module = rel_module( m, addr, &len, function, (size_t *)&func_addr );
	free( module );
	if ( func_addr < STACK ) {
		fatal( "Function not found !\n" );
		return;
	}

	address_save = khijack_syscall( SYSCALL_KFREE, func_addr, 
		file_descriptor_kmem );
	if ( address_save < 0 ) {
		return;
	}

	call_syscall( SYSCALL_KFREE );
	n = khijack_syscall( SYSCALL_KFREE, address_save, file_descriptor_kmem );
	if ( n < 0 ) {
		return;
	}

	/* So the module could be freed now */
	kfree( addr );
	return;
}

int
kinsert( char * m, size_t location, char * function, size_t * addr )
{
	void * module;
	int len;
	int n;

	module = rel_module( m, location, &len, function, addr );
	if ( !module ) { 
		fatal( "cannot make relocation for %s..\n\n", m );
		return( -1 );
	}

	n = write_kmem( module, location, len, file_descriptor_kmem );
	if ( n < 0 ) {
		fatal( "error, module %s NOT inserted into /dev/kmem..\n\n", m );
		return( -1 );
	}	

	fprintf( stdout, "module %s inserted in /dev/kmem..\n", m );
	return( 0 );
}

void
kscan( size_t lower, size_t upper, char * string, int fd )
{
	char buf[ BUFSIZ ], *mm = NULL;
	size_t addr;
	int l = strlen( string );
	int j = 0;
	int p = 0;
	int n;
	int max;
	int len;
	size_t prev = 0x0;

	bzero( buf, BUFSIZ );
	max = BUFSIZ;
	fputs( "scanning all kernel data... wait :\n\n", stdout );

	for ( addr = lower; addr < upper; addr += (BUFSIZ-l) ) {
		n = read_kmem( &buf, addr, BUFSIZ, fd );
		if ( n == 0 ) {
			len = 0;
	
			do {
				mm = memmem( buf+len+l, BUFSIZ-len, string, l ) ;
				len = mm - buf ;
				if ( addr + len > upper ) {
					break;
				}

				if ( mm != NULL ) {
					fprintf( stdout, "pattern found at %#x ", (size_t)(mm-buf)+addr );
					prev = (size_t)( addr+len+l );
					fprintf( stdout, "\n" );
					if ( !script_on_command_line ) {
						j++;
						if ( j == 23 ) {
							char c = enter();
							if ( c == 'q' ) {
								return;
							}	
							j = 0;
						}	
					}
					fflush( stdout );
					n++;
				}
			} while ( mm != NULL );

			if ( n ) {
				p += n;
			}
		}
	}

	if ( p == 0 ) {
		fatal( "no pattern found\n" );
	} else {
		fprintf( stdout, "found %d pattern \n", p );
	}
	return;
}

void
kcopy( size_t source, size_t destination, int ssize )
{
	char * cache = NULL;
	int n;
	
	cache = malloc( ssize+1 );
	if ( cache == NULL ) {
		perror( "kcopy:malloc()" );
		return;
	}
	
	n = read_kmem( cache, source, ssize, file_descriptor_kmem );
	if ( n == 0 ) {
		cache[ssize+1] = '\0';
		n = write_kmem( cache, destination, ssize, file_descriptor_kmem );
		if ( n < 0 ) {
			return;
		}
	}
	free( cache );
	return;
}

void
kwrite( size_t location, char * string )
{
	char * str2hex = NULL;
	char c;
	int i;
	int n;
	
	str2hex = string2hex( string, &i );
	if ( str2hex == NULL ) {
		return;
	}

	fputs( "\n[ before ]\n", stdout );
	kdump( location, i, file_descriptor_kmem );
	fputs( "\n[ after ]\n", stdout );
	hexdump( str2hex, i, location );
	fputs( "\nreally write [y/n] ? ", stdout );
	c = getchar();
	getchar();
	if ( c == 'y' || c == 'Y' ) {
		n = write_kmem( str2hex, location, i, file_descriptor_kmem );
		if ( n == 0 ) {
			fprintf( stdout, "string %s written in /dev/kmem\n", 
				string );
		}
	} else {
		fatal( "don't write in /dev/kmem..\n" );
	}
	return;
}

void
usage_kmem( void )
{
    fputs( 
	"kmem [ -mfdiscwkugrno ] <arguments>\n\n"
	"This command makes it possible to handle the device /dev/kmem\n"
	"with its own way, to read there, to write there, and mainly to\n"
	"include there files objects.\n\n"
	"__options\n"
	"  -m [size] : allocate memory capacity /dev/kmem ;\n"
	"  -f [address] : free an allocated memory capacity ;\n"
	"  -d [address] [size] [file] : display a block of file\n" 
	"<file> (/dev/kmem ou /dev/mem) according to the offset and the size ;\n"
	"  -g [address] [size] [file] : dissasemble this block ;\n"
	"  -i [module] [address] <symbol_to_get_addr>: insert a module in \n"
	"/dev/kmem at address <address> ;\n"
	"  -k [module] [function] : insmod <module> in /dev/kmem\n"
	"<function> will be executed (commonly init_module) ;\n"
	"  -r [module] [module_address] [function] : rmmod [module_address]\n"
	"and calling before [function] (commonly cleanup_module) ;\n"
	"  -s [start address] [end address] [string] [file] : search a string\n"
	"in <file> (/dev/kmem ou /dev/mem) ;\n"
    "  -o [start address] [end address] [opcode] [file] : search opcode\n"
	"in <file> (/dev/kmem ou /dev/mem) ;\n"
	"  -u : toggle the use of mmap_kmem or write_kmem ;\n"
	"  -c [source address] [destination address] [size] : copy a block\n"
	"of /dev/kmem from an address towards another ;\n"
	"  -w [address] [string] : write a string in /dev/kmem\n"
	"(e.g. -w 0xc0000000 \\x41\\x42\\x43\\x44).\n\n", stdout );
    return;
}	

enum {
	WRITE_IN_KMEM,
	DUMP_KMEM,
	INSERT_IN_KMEM,
	SCAN_KMEM,
	COPY_KMEM,
	INSMOD_IN_KMEM,
	RMMOD_IN_KMEM,
	DISS_KMEM
};

void
do_kmem( int pargc, char * pargv[] )
{
	char * file = NULL;
	char c;
	size_t l;
	size_t ll;
	int ssize;
	int flag;
	int file_descriptor_mem;
	int kma;

	NOARGUMENTS( pargc, 1, kmem );

	ssize = 0;
	flag = -1;
	file = NULL;
	optind = 0;

	while ( (c = getopt(pargc, pargv, "uidcskwgrom:f:")) != EOF ) {
		switch ( c ) {
		case 'u':
			if ( mmap_use > 0 ) {
				mmap_use--;
			} else {
				mmap_use++;
			}
			break;
		case 'm':
			ssize = atoi( optarg );
			/*
			 if ( ssize < 0 || ssize > 1024 ) {
			 	fatal( "size must be > 0 and < 1024..\n" );
			 	return;
			 }
			 */
			kmalloc( ssize, (unsigned long *)&kma );
			break;
		case 'f':
			XSTRTOUL( optarg, l );
			kfree( l );
			break;
        case 'w':
            flag = WRITE_IN_KMEM;
            break;
        case 'd':
            flag = DUMP_KMEM;
            break;
        case 'i':
            flag = INSERT_IN_KMEM;
            break;
        case 'k':
            flag = INSMOD_IN_KMEM;
            break;
		case 'r':
			flag = RMMOD_IN_KMEM;
			break;
        case 's':
            flag = SCAN_KMEM;
            break;
        case 'c':
            flag = COPY_KMEM;
            break;
		case 'o':
			pargv[4] = string2hex( pargv[4], &ssize );
			if ( pargv[4] == NULL ) {
				return;
			}
			flag = SCAN_KMEM;
			break;
		case 'g':
			flag = DISS_KMEM;
			break;
        default:
            usage_kmem();
            return;
        }
    }	

    switch ( flag ) {
    case WRITE_IN_KMEM:
        if ( pargc != 4 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        kwrite( l, pargv[3] );
        break;

    case DUMP_KMEM:
        if ( pargc < 4 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        ssize = atoi( pargv[3] );
			
		fprintf( stdout, "\n[ %#x -> %#x ]\n", l, l+ssize );
        switch ( pargc ) {
        case 4:
            kdump( l, ssize, file_descriptor_kmem );
            break;
        case 5:
            file_descriptor_mem = open( pargv[4], O_RDONLY );
            if ( file_descriptor_mem < 0 ) {
                fatal( "can't open %s\n", pargv[4] );
                return;
            }

            kdump( l, ssize, file_descriptor_mem );
        }
        break;

	case DISS_KMEM:
		if ( pargc < 4 ) {
			usage_kmem();
			return;
		}

		XSTRTOUL( pargv[2], l );
		ssize = atoi( pargv[3] );

		fprintf( stdout, "\n[ %#x -> %#x ]\n", l, l+ssize );
		switch ( pargc ) {
		case 4:
			kdisass( l, ssize, file_descriptor_kmem );
			break;
		case 5:
			file_descriptor_mem = open( pargv[4], O_RDONLY );
			if ( file_descriptor_mem < 0 ) {
				fatal( "Can't open %s\n", pargv[4] );
				return;
			}
			kdisass( l, ssize, file_descriptor_mem );
		}
		break;

    case INSERT_IN_KMEM:
        if ( pargc < 4 || pargc > 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[3], l );
		switch (pargc) {
			case 4:
		        kinsert( pargv[2], l, 0, 0 );
				break;
			case 5:
				kinsert( pargv[2], l, pargv[4], &l );
				fprintf( stdout, "%s located at %#08x\n", pargv[4], l );
				break;
			default:
				break;
		}
        break;

	case RMMOD_IN_KMEM:
		if ( pargc != 5 ) {
			usage_kmem();
			return;
		}
		XSTRTOUL( pargv[3], l );
		krmmod( pargv[2], l, pargv[4] );
		break;

    case INSMOD_IN_KMEM:
        if ( pargc != 4 ) {
            usage_kmem();
            return;
        }

        kinsmod( pargv[2], pargv[3] );
        break;

    case SCAN_KMEM:
        if ( pargc < 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        XSTRTOUL( pargv[3], ll );
		if ( ll - l < 0 ) {
			fatal( "size of memory area you want to scan "
				"is negative !\n" );
			return;
		}

        switch ( pargc ) {
        case 5:
           	kscan( l, ll, pargv[4], file_descriptor_kmem );
            break;
        case 6:
            file_descriptor_mem = open( pargv[5], O_RDONLY );
            if ( file_descriptor_mem < 0 ) {
                fatal( "can't open %s\n", pargv[5] );
                return;
            }
            kscan( l, ll, pargv[4], file_descriptor_mem );
        }
        break;

    case COPY_KMEM:
        if ( pargc != 5 ) {
            usage_kmem();
            return;
        }

        XSTRTOUL( pargv[2], l );
        XSTRTOUL( pargv[3], ll );
        kcopy( l, ll, atoi(pargv[4]) );
        break;

    default:
        return;
    }

    return;
}

