/* 
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <errno.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "../libdisasm/libdis.h"
#include "include/kernsh.h"

struct new_module_symbol {
    size_t value;
    unsigned long name;
};

size_t
get_symbol_by_query_module( char * symbol )
{
	struct new_module_symbol * syms = NULL;
	struct new_module_symbol * s = NULL;
	size_t len_new_module_symbol;
	size_t ret;
	size_t j;
	int n;

	len_new_module_symbol = sizeof( struct new_module_symbol ) - 1;
	syms = malloc( len_new_module_symbol );
	if ( syms == NULL ) {
		perror( "get_symbol_by_query_module:malloc()" );
		return( 0 );
	}

retry :
	n = query_module( NULL, QM_SYMBOLS, syms, len_new_module_symbol, 
	&ret );
	if ( n == -1 ) {
		if ( errno == ENOSPC ) {
			len_new_module_symbol = ret;
			syms = ( struct new_module_symbol * )realloc( syms, 
			len_new_module_symbol );
			goto retry;
		}
		else {
			free( syms );
			return( 0 );
		}
	}

	for ( j = 0, s = syms; j < ret; ++j, ++s ) {
		if ( !strncmp((char *)syms+s->name, symbol, 
		strlen((char *)syms+s->name)) ) {
			ret = s->value;
			free( s );
			return( ret );
		}
	}

	free( syms );
	return( 0 );
}

/* Obsolete function */
#define MAX_SYMS 4096
size_t
get_symbol_by_kernel_sym( char * symbol )
{
	struct kernel_sym tab_kernel_sym[ MAX_SYMS*2 ];
	int numsyms;
	int i;

	/* From kernel 2.4.19 sources 
	 *
	 * get_kernel_syms() :
	 * Copy the kernel symbol table to user space.  If the argument is
	 * NULL, just return the size of the table.
	 *
	 * This call is obsolete.  New programs should use 
	 * query_module+QM_SYMBOLS which does not arbitrarily limit the 
	 * length of symbols.
	 */
	numsyms = get_kernel_syms( NULL );
	if ( numsyms < 0 || numsyms > MAX_SYMS ) {
		return( 0 );
	}

	get_kernel_syms( tab_kernel_sym );
	for ( i = 0; i < numsyms; i++ ) {
		if ( !strncmp(symbol, tab_kernel_sym[i].name, strlen(symbol)) ) {
			return( tab_kernel_sym[i].value );
		}
	}
	return( 0 );
}

#define BOOT "/boot"
#define SYSBOOT "/System.map-"
#define SRC "/usr/src/linux-"
#define SYSTEM "/System.map"

FILE *
found_system_map( void )
{
	FILE * fd_file_system_map = NULL;
	struct utsname kernel;
	char * file_system_map = NULL;
	char * p_file_system_map = NULL;
	int len_file_system_map;

	if ( !path_system_map ) {
		uname( &kernel );

		len_file_system_map = strlen( kernel.release ) + strlen( BOOT ) 
			+ strlen( SYSBOOT );
		file_system_map = malloc( len_file_system_map+1 );
		if ( file_system_map == NULL ) {
			perror( "found_system_map:malloc()" );
			return( NULL );
		}
		p_file_system_map = file_system_map;
	
		memcpy( p_file_system_map, BOOT SYSBOOT, 
			strlen(BOOT)+strlen(SYSBOOT) );
		p_file_system_map += strlen( BOOT ) + strlen( SYSBOOT );
		memcpy( p_file_system_map, kernel.release, strlen(kernel.release) );
		p_file_system_map += strlen( kernel.release );
		*p_file_system_map = '\0';

		fd_file_system_map = fopen( file_system_map, "r" );
		if ( !fd_file_system_map ) {
			free( file_system_map );
			fd_file_system_map = fopen( "/boot/System.map", "r" );
			if ( !fd_file_system_map ) {
				fd_file_system_map =
				fopen( "/usr/src/linux/System.map", "r" );
				if ( !fd_file_system_map ) {
					len_file_system_map = strlen( kernel.release ) + 
						strlen( SRC ) + strlen( SYSTEM );
					file_system_map = malloc( len_file_system_map+1 );
					if ( file_system_map == NULL ) {
						perror( "found_system_map:malloc()" );
						return( NULL );
					}
					p_file_system_map = file_system_map;

					memcpy( p_file_system_map, SRC, strlen(SRC) );
					p_file_system_map += strlen( SRC );
					memcpy( p_file_system_map, kernel.release, 
						strlen(kernel.release) );
					p_file_system_map += strlen( kernel.release );
					memcpy( p_file_system_map, SYSTEM, strlen(SYSTEM) );
					p_file_system_map += strlen( SYSTEM );
					*p_file_system_map = '\0';

					fd_file_system_map = fopen( file_system_map, "r" );
					free( file_system_map );
				}
			}
		}
	} else {
		fd_file_system_map = fopen( path_system_map, "r" );
	}

	return( fd_file_system_map );
}

#define VMLINUX "/vmlinux"

int
found_vmlinux( void )
{
	struct utsname kernel;
	char * file_vmlinux = NULL;
	char * p_file_vmlinux = NULL;
	int len_file_vmlinux;
	int fd_file_vmlinux;

	if ( !path_vmlinux ) {
		uname( &kernel );

		len_file_vmlinux = strlen( SRC ) + strlen( kernel.release ) + 
			strlen( VMLINUX );
		file_vmlinux = malloc( len_file_vmlinux+1 );
		if ( file_vmlinux == NULL ) {
			perror( "found_vmlinux:malloc()" );
			return( -1 );
		}
		p_file_vmlinux = file_vmlinux;
	
		memcpy( p_file_vmlinux, SRC, strlen(SRC) );
		p_file_vmlinux += strlen( SRC );
		memcpy( p_file_vmlinux, kernel.release, strlen(kernel.release) );
		p_file_vmlinux += strlen( kernel.release );
		memcpy( p_file_vmlinux, VMLINUX, strlen(VMLINUX) );
		p_file_vmlinux += strlen( VMLINUX );
		*p_file_vmlinux = '\0';

		fd_file_vmlinux = open( file_vmlinux, O_RDONLY );
		if ( fd_file_vmlinux < 0 ) {
			free( file_vmlinux );
			len_file_vmlinux = strlen ( BOOT ) + strlen ( VMLINUX ) 
				+ /* - */ 1 + strlen( kernel.release );
			file_vmlinux = malloc( len_file_vmlinux+1 );
			if ( file_vmlinux == NULL ) {
				perror( "found_vmlinux:malloc()" );
				return( -1 );
			}
			p_file_vmlinux = file_vmlinux;

			memcpy( p_file_vmlinux, BOOT, strlen(BOOT) );
			p_file_vmlinux += strlen ( BOOT );
			memcpy( p_file_vmlinux, VMLINUX "-", 
				strlen(VMLINUX)+strlen("-") );
			p_file_vmlinux += strlen( VMLINUX ) + strlen( "-" );
			memcpy( p_file_vmlinux, kernel.release, 
				strlen(kernel.release) );
			p_file_vmlinux += strlen ( kernel.release );
			*p_file_vmlinux = '\0';

			fd_file_vmlinux = open( file_vmlinux, O_RDONLY );
			if ( fd_file_vmlinux < 0 ) {
				free( file_vmlinux );

				fd_file_vmlinux = open( "/usr/src/linux/vmlinux", O_RDONLY );
				if ( fd_file_vmlinux < 0 ) {
					fatal( "no such file vmlinux..\n" );
				}
			}
		}

		if ( file_vmlinux != NULL ) {
			free ( file_vmlinux );
		}
	} else {
		fd_file_vmlinux = open( path_vmlinux, O_RDONLY );
	}
	return( fd_file_vmlinux );
}

int 
firstspace( char * c )
{
	int i;
	
	for ( i = strlen(c); i; i-- ) {
		if ( c[i] == ' ' ) {
			return( i );
		}
	}
	return( 0 );
}

void
build_address( char * address, char * line )
{
	memcpy( address, "0x", 2 );
	address += 2;
	if ( !strstr(line, "00000000") ) {
		memcpy( address, line, 8 );
		address += 8;
	}
	else {
		memcpy( address, line+8, 8 );
		address += 8;
	}
	*address = '\0';
}

size_t
get_symbol_by_system_map( char * symbol, int regexp )
{
	FILE * fd_file_system_map;
	char line[ 128 ];
	char * address = NULL;
	int len_address;
	size_t ret = 0;
	int i = 0;

	fd_file_system_map = found_system_map();
	if ( fd_file_system_map == NULL ) {
		return( 0 );
	}

	len_address = 2 * sizeof( size_t ) + 2;
	address = malloc( len_address+1 );
	if ( address == NULL ) {
		perror( "get_symbol_by_system_map:malloc()" );
		return( 0 );
	}
	memset( address, '\0', len_address+1 );

	for ( ; fgets(line, sizeof(line)-1, fd_file_system_map) != 
		NULL ; ) {
		while ( strlen(line) && (line[strlen(line)-1] == '\n' ||
		line[strlen(line)-1] == '\r') ) {
			line[strlen(line)-1] = '\0';
			if ( !regexp ) {
				if ( !strcmp(&line[firstspace(line)+1], symbol) ) {
					build_address( address, line );
					fclose( fd_file_system_map );
					ret = strtoul( address, NULL, 16 );
					free( address );
					return( ret );
				}
			}
			else {
				if ( strstr(&line[firstspace(line)+1], symbol) ) {
					build_address( address, line );
					if ( regexp == 1 ) {
						if ( !i ) {
							fprintf( stdout, "\n%-30s\t\t%s\n\n", "SYMBOL", "ADDRESS" );
							i++;
						}
						fprintf( stdout, "%-30s\t\t%s\n", 
							&line[firstspace(line)+1], address );
					} else {
						if ( line[firstspace(line)+1] != '_' )
						if ( strtoul(address, NULL, 16) > regexp ) {
							ret = strtoul(address, NULL, 16);
							free( address );
							return( ret );
						}
					}
				}
			}
		}
	}

	free( address );
	fclose( fd_file_system_map );

	if ( regexp && !i ) {
		fprintf( stdout, "no symbol matching this regexp\n" );
	}
	return( 0 );
}

size_t
get_symbol_by_name( char * symbol )
{
	size_t p_symbol = 0;
	int regexp;

	regexp = 0;
	if ( symbol != NULL ) {
    	p_symbol = get_symbol_by_kernel_sym( symbol );
    	if ( !p_symbol ) {
        	p_symbol = get_symbol_by_query_module( symbol );
        	if ( !p_symbol ) {
				p_symbol = get_symbol_by_system_map( symbol, regexp );
				if ( !p_symbol ) {
					return( 0 );
				}
			}
		}
	}
	return( p_symbol );
}

/* we use only map file contrarily to get_symbol_by_name() */ 
char *
get_symbol_by_addr( char * address )
{
	FILE * fd_file_system_map;
	char line[ 128 ];
	char * name_symbol = NULL;
	char * p_address = NULL;
	int len_name_symbol;

	fd_file_system_map = found_system_map();
	if ( fd_file_system_map == NULL ) {
		return( NULL );
	}

	p_address = &address[ 2 ];

	for ( ; fgets(line, sizeof(line)-1, fd_file_system_map) != 
	NULL ; ) {
		while ( strlen(line) && (line[strlen(line)-1] == '\n' ||
		line[strlen(line)-1] == '\r') ) {
			line[strlen(line)-1] = '\0';
			if ( strstr(line, p_address) ) {
				len_name_symbol = 
				strlen( &line[firstspace(line)+1] );

				name_symbol = malloc( len_name_symbol+1 );
				if ( name_symbol == NULL ) {
					perror( "get_symbol_by_addr:malloc()" );
					return( NULL );
				}
				memset( name_symbol, '\0', len_name_symbol+1 );

				memcpy( name_symbol, &line[firstspace(line)+1], 
					len_name_symbol );
				fclose( fd_file_system_map );
				return( name_symbol );
			}
		}
	}

	fclose( fd_file_system_map );
	return( address );
}

size_t
get_next_symbol_by_addr( size_t address )
{
	FILE * fd_file_system_map;
	char line[ 128 ];
	char * p_address = NULL;
	char * next_address = NULL;
	int len_p_address;
	int len_next_addr; 

	len_p_address = 2 * sizeof( size_t ) + 2;
	p_address = malloc( len_p_address+1 );
	if ( p_address == NULL ) {
		perror( "get_next_symbol_by_addr:malloc()" );
		return( 0 );
	}

	memset( p_address, '\0', len_p_address+1 );
	snprintf( p_address, len_p_address+1, "%08x", address );

	fd_file_system_map = found_system_map();
	if ( fd_file_system_map == NULL ) {
		return( -1 );
	}

	for ( ; fgets(line, sizeof(line)-1, fd_file_system_map) != 
		NULL ; ) {
		while ( strlen(line) && (line[strlen(line)-1] == '\n' ||
		line[strlen(line)-1] == '\r')) {
			line[strlen(line)-1] = '\0';
			if ( strstr(line, p_address) ) {
				fgets( line, sizeof(line)-1, fd_file_system_map );
				while ( strlen(line) && (line[strlen(line)-1] == '\n' ||
				line[strlen(line)-1] == '\r')) {
					line[strlen(line)-1] = '\0';
					
					len_next_addr = 2 * sizeof(size_t) + 2;
					next_address = malloc( len_next_addr+1 );
					if ( next_address == NULL ) {
						perror( "get_next_symbol_by_addr:malloc()" );
						return( 0 );
					}
					memset( next_address, '\0', len_next_addr+1 );

					build_address( next_address, line );
					fclose( fd_file_system_map );
					return( strtoul(next_address, NULL, 16) );
				}
			}
		}
	}

	free( p_address );
	free( next_address );	
	fclose( fd_file_system_map );
	return( 0 );
}

size_t
display_desc_symbol( char * symbol, size_t start_address )
{
	size_t end_address = 0;

	fprintf( stdout, "       symbol : %s\n", symbol );
	fprintf( stdout, "start address : %#x\n", start_address );

	end_address = get_next_symbol_by_addr( start_address );
	if ( end_address != 0 ) {
		fprintf( stdout, "  end address : %#x\n", end_address );
		fprintf( stdout, "       length : %d\n\n", 
			end_address-start_address );
	}
	return( end_address );
}

void
display_desc_symbol_with_regexp( char * symbol )
{
	int regexp = 1;

	/* we use only get_symbol_by_system_map() with regexp but later
	 * we'll must use get_symbol_by_name()
	 */
	get_symbol_by_system_map( symbol, regexp );
	return;
}

void
usage_symbol( void )
{
    fputs(
    	"symbol [ -naq ] <arguments>\n\n"
    	"This command makes it possible to display the description of a\n"
    	"symbol according to whether we provide a name, an address or a\n"
    	"regular expression..\n\n"
    	"__options\n"
    	"  -n [name] <opcode|disass> : display characteristics of a symbol\n"
    	"according to the name ;\n"
    	"  -a [address] <opcode|disass> : display characteristics of a\n"
    	"symbol according to the address ;\n"
    	"  -q [regexp] : display a list of symbols according to a regular\n"
    	"expression.\n\n", stdout );
	return;
}

void
do_symbol( int pargc, char * pargv[] )
{
	char * symbol = NULL;
	char c;
	size_t start_address = 0;
	size_t end_address = 0;
	/* int i; */

	NOARGUMENTS( pargc, 1, symbol );

	optind = 0;
	while ( (c = getopt(pargc, pargv, "n:a:q:")) != EOF ) {
        switch ( c ) {
		case 'n':
			start_address = get_symbol_by_name( optarg );
			if ( !start_address ) {
				fatal( "can't find symbol %s..\n", optarg );
				return;
			}

			end_address = display_desc_symbol( optarg, start_address );
			break;

		case 'a':
			XSTRTOUL( optarg, start_address	);

			symbol = get_symbol_by_addr( optarg );
			if ( symbol == NULL ) {
				fatal( "can't find symbol to address %s..\n", 
					optarg );
				return;
			}

			end_address = display_desc_symbol( symbol, start_address );
			if ( symbol != optarg ) {
				free ( symbol );
			}

			break;
				
		case 'q':
			display_desc_symbol_with_regexp( optarg );
			break;

		default:
			usage_symbol();
			return;
	}
	}

	if ( pargv[3] && end_address != 0 ) {
		if ( !strcmp(pargv[3], "opcode") ) {
			fprintf( stdout, "[!] opcode from address %#x..\n\n", 
				start_address );
			kdump( start_address, end_address-start_address,
				file_descriptor_kmem );
		}

		if ( !strcmp(pargv[3], "disass") ) {
			fprintf( stdout, "[!] disassembling from address %#x..\n\n",
				start_address );
			/*
			i = disasm( start_address, end_address-start_address );
			if ( i < 0 ) {
				fatal( "can't disass this symbol..\n" );
			}
			*/
			kdisass( start_address, end_address-start_address, 
				file_descriptor_kmem );
		}
		start_address = end_address = 0;
	}
	return;
}

